/* design rule check. The main body */


#include <stdio.h>
#include <stdarg.h>

#include <stdlib.h>
#include <strings.h>

#include "d.h"


#define DEFAULT_ATTR_FILE "pdr"

ATTRDEF  attribute[MAXATTR];
int      num_of_attr;
char     *typex_name;
ATTRNODE layers[MAXLAYER];
int      num_te_errs = 0;
int      num_te_msgs = 0;

CONDVEC truevec;             /* all values DRC_TRUE */
CONDVEC current_conditions;  /* temp variable */

STACK  *top;
CSTACK *ctop;

FILE *attdrcin;            /* file pointer to attribute file */
FILE *tedrcin;             /* file pointer to type expressions file */
extern int attdrcdebug;
extern int tedrcdebug;
extern int attdrc_flex_debug;
extern int tedrc_flex_debug;    

int aflag;                /* boolean flag to print out attribute data */
int oflag;                /* boolean flag to print out output of type expression */

int currlayer = -1;       /* current index to "layers[]" */


/* local procedures */

static void writeout(void);
static void error(char *format, ...);
static void errorhead(int);
static char *value_word(DRV i);
static void pr_usage(char *fname);

/************************
  attribute side 
************************/


/* add another attribute and its description to the arrays */
void addattr(char *name, char *desc, DRV initvalue)
{ 
    int i = num_of_attr;
    stralloc(attribute[i].name,name);
    /* must trim surrounding quotes off of desc string */
    desc[strlen(desc)-1] = 0;
    stralloc(attribute[i].desc,desc+1);
    attribute[i].initvalue = initvalue;
    num_of_attr++;
}

/* add another layer name entry */
void add_to_lnames(char *s)
{
    int i = currlayer;
    int j,k;
    i++; 
    if ( i == MAXLAYER ) { 
        error("Sorry. Exceeded maximum number of layers %d",MAXLAYER);
        exit(EXIT_FAILURE);
    }

    stralloc(layers[i].name, s);

    /* initialize all the attributes */
    for(j=0;j<num_of_attr;j++) { 
        layers[i].pre[j]   = DRC_TRUE;
        layers[i].wpre[j]  = DRC_TRUE;
        layers[i].postr[j] = DRC_INHERIT;
        for(k=0;k<MAXPARAM;k++) { 
            layers[i].post[k][j]  = DRC_INHERIT;
            layers[i].prer[k][j]  = DRC_TRUE;
            layers[i].wprer[k][j] = DRC_TRUE;
        }
        /* NOTE:  if mergeval of layer is true, then current
                  conditions are propagated to each parameter.
                  Otherwise, attributes are set to their initial
                  conditions and are inherited (if not overwritten). 
           NOTE:  postconditions for all parameters are initially
                  set to inherit */
    }
    layers[i].num_param = 0;
    layers[i].realmtype = 0;
    for(k=0;k<MAXPARAM;k++)
        layers[i].param_rtype[k] = 0;
    currlayer = i;
}

/* 
   set attribute qualifiers. If "copy" value is set, 
   then copy DRC_TRUE to all postconditions 
*/
void setattrqual(BOOL mval, BOOL cpval)
{
    int i,j;

    layers[currlayer].mergeval = mval;
    layers[currlayer].copyval = cpval;
    if ( cpval ) { 
        /* copy the initial data over now */
        for(j=1;j<MAXPARAM; j++)  
            for(i=0;i<MAXATTR; i++)  
                layers[currlayer].post[j][i] =
                   layers[currlayer].post[0][i];
    }
}

/* set the realm type for the component */
void set_realm_type(char *realmname)
{
    stralloc(layers[currlayer].realmtype, realmname);
}

/* add the realm type of the child to the array of parameter realm types */
void addrparam(char *rp_name)
{
    int i = layers[currlayer].num_param;
    stralloc(layers[currlayer].param_rtype[i],rp_name); 
    layers[currlayer].num_param++;
}

/* set the attribute for the current component 
   (1) prop is the name of the property
   (2) a_type is the attribute type (whether "assert" or "unset" or ...)
   (3) ctype is the condition type (whether "precondition" or "postrestriction" or ... )
   (4) param is the index to the condition, if the condition takes indices
   (5) wtype is true if the condition is flagged as a warning, rather than an error.
*/

void do_props(char *prop, int a_type, int ctype, int param, int wtype)
{
    int i; 
    int j;
    int found = 0;

    for(i=0;i<num_of_attr && !found ;i++) { 
        if ( !strcmp(attribute[i].name, prop)) { 
            /* then i is the location of the property */
            found = 1;
            if ( aflag ) 
                printf("%s : %s : %d %d %d %d \n",layers[currlayer].name, prop, a_type, ctype, param, wtype);

            /* check to see if the postXXXX command is followed by a proper 
               attribute */
            if ( (ctype==2)||(ctype==3)||(ctype==6)||(ctype==7)) { 
                if (!( (a_type==DRC_ASSERT)||(a_type==DRC_NEGATE)||(a_type==DRC_UNSET))) { 
                      error("Postcondition/Postrestrictions can only be used with\n"\
                            " attributes \"assert\", \"negate\", or \"unset\".");
                      break;
                  }
            }


            if ( !wtype ) { 
                switch (ctype) {
                  case 1: layers[currlayer].pre[i] = a_type; break;
                  case 2: layers[currlayer].post[param-1][i] = a_type; break;
                  case 3: 
                    if (layers[currlayer].copyval) {
                        for(j=0;j<MAXPARAM;j++) layers[currlayer].post[j][i] = a_type; 
                    } else { 
                        layers[currlayer].post[0][i] = a_type;
                    }
                    break;
                  case 4: layers[currlayer].prer[param-1][i] = a_type; break;
                  case 5: for(j=0;j<MAXPARAM;j++) layers[currlayer].prer[j][i] = a_type; break; 
                  case 6: layers[currlayer].postr[i] = a_type; break;
                  case 7:
                    if (layers[currlayer].copyval) {
                        for(j=0;j<MAXPARAM;j++) layers[currlayer].post[j][i] = a_type; 
                    } else { 
                        layers[currlayer].post[0][i] = a_type;
                    }
                    layers[currlayer].postr[i] = a_type; 
                    break;
                }
            } else /* warning types */ { 
                switch (ctype) {
                  case 1: layers[currlayer].wpre[i] = a_type; break;
                  case 2: error("postconditions can't have warnings."); break;
                  case 3: error("postconditions can't have warnings."); break;
                  case 4: layers[currlayer].wprer[param-1][i] = a_type; break;
                  case 5: for(j=0;j<MAXPARAM;j++) layers[currlayer].wprer[j][i] = a_type; break;
                  case 6: error("postrestrictions can't have warnings."); break;
                  case 7: error("postrestrictions and postconditions can't have warnings."); break;
                }
            }
        }
    }
    if ( !found ) error ("property %s not found",prop);
}
                
/* write out the attribute data to standard output */
static void writeout(void)
{
    ATTRNODE a;
    int i;
    int j;
    int k;
    int l;

    printf("#define MAXATTR %d\n",MAXATTR);
    printf("#define MAXPARAM %d\n",MAXPARAM);
    printf("#define MAXLAYER %d\n",MAXLAYER);
    printf("int attrlen = %d;\n",num_of_attr);
    printf("int num_of_layers = %d;\n",currlayer);

    printf("char *attrname[] = { ");
    for(i=0; i<num_of_attr; i++)
        printf("\"%s\", ",attribute[i].name);
    printf("\"\" };\n");

    printf("char *attrdesc[] = { ");
    for(i=0; i<num_of_attr; i++)
        printf("%s, ",attribute[i].desc);
    printf("\"\" };\n");

    printf("ATTRNODE *attrlist[] = {\n");
    for(k = 0; k<=currlayer; k++) {
        a = layers[k];
        printf("\n { \"%s\", %d, %d, \"%s\", %d, ",
               a.name, a.copyval, a.mergeval, a.realmtype, a.num_param);

        printf(" { ");
        for(i=0; i<MAXPARAM; i++) {
            if ( i < a.num_param) { 
                printf("\"%s\"",a.param_rtype[i]);
            } else { 
                printf("\"\"");
            }
            if ( i < MAXPARAM-1 ) { 
                printf(", ");
            } else { 
                printf(" }, \n");
            }
        }

        printf("   { ");
        for(i=0; i<MAXATTR-1; i++)
          printf("%d, ",a.pre[i]);
        printf("%d },\n",a.pre[MAXATTR-1]);

        printf("   { ");
        for(i=0; i<MAXATTR-1; i++)
          printf("%d, ",a.wpre[i]);
        printf("%d },\n",a.wpre[MAXATTR-1]);

        printf("   {\n");
        for(j=0; j<MAXPARAM; j++) { 
            printf("      { ");
            for(i=0; i<MAXATTR-1; i++)
              printf("%d, ",a.post[j][i]);
            printf("%d },\n",a.post[j][MAXATTR-1]);
        }

        printf("   },\n");
        printf("   {\n");
        for(j=0; j<MAXPARAM; j++) { 
            printf("      { ");
            for(i=0; i<MAXATTR-1; i++)
              printf("%d, ",a.prer[j][i]);
            printf("%d },\n",a.prer[j][MAXATTR-1]);
        }
        printf("   },\n");

        printf("   },\n");
        printf("   {\n");
        for(j=0; j<MAXPARAM; j++) { 
            printf("      { ");
            for(i=0; i<MAXATTR-1; i++)
              printf("%d, ",a.wprer[j][i]);
            printf("%d },\n",a.wprer[j][MAXATTR-1]);
        }
        printf("   },\n");

        printf("   { ");
        for(i=0; i<MAXATTR-1; i++)
          printf("%d, ",a.postr[i]);
        printf("%d },\n",a.postr[MAXATTR-1]);

        printf(" },");
    }
    printf("\n};\n\n");
}

/*****************************************************

  THE TYPE EXPRESSION SIDE OF THINGS 

*****************************************************/

/********** debugging routines - call from debugger ****/

void printc( CONDVEC *x )
{
   int i;

   printf("\n");
   foreach_attribute(i) {
      printf("%s %s", value_word((*x)[i].cpred), attribute[i].name);
      if ((*x)[i].origin)
         printf(" from %s", (*x)[i].origin->name);
      printf("\n");
   }
}

void printp( DRV array[] )
{
    int i;

    printf("\n");
    foreach_attribute(i) {
      printf("%s %s\n", value_word(array[i]), attribute[i].name);
    }
}
 
/********** COND routines *******************/

void copycond( CONDVEC *x, CONDVEC *y )
{
   int i;

   foreach_attribute(i) {
      (*x)[i].origin = (*y)[i].origin;
      (*x)[i].cpred  = (*y)[i].cpred;
   }
}

void setcpostres( CONDVEC *x, TEX *t )
{
   int i;

   foreach_attribute(i) {
      (*x)[i].cpred  = t->attrdata->postr[i];
      (*x)[i].origin = t;
   }
}


void init_current_conditions( TEX *t )
{
   int i;

   foreach_attribute(i) {
      current_conditions[i].cpred  = attribute[i].initvalue;
      current_conditions[i].origin = t;
   }
}

/* print message if this is the first error in this type expression */
static void errorhead(int err_or_warn)
{
    if (err_or_warn == ERR_MSG)
       num_te_errs++;
    if ( num_te_msgs++ == 0 ) 
        printf("=============\nIn type expression %s:\n",typex_name);
}

/* set the type expression name */
void set_typex_name(char *name)
{
    stralloc(typex_name, name);
}

/* print the attribute values of the typex expression t */
void print_top_te_values(TEX *t)
{
    int i;
    if ( oflag ) { 
        printf("**Value of system root for %s:\n",typex_name);
        for(i=0;i<num_of_attr;i++) { 
            printf("   %20s  %s\n",
                   attribute[i].name, value_word(t->cpostres[i].cpred));
        }
        printf("\n");
    }
}


char rumpname[MAXLENGTH];

void truncate( char *name )
{
   int  i, j;

   strcpy(rumpname, name);
   i = strlen(name) - 1;
   while (*(name+i) >= '0' && *(name+i) <= '9')
      i--;
   for (j=0; j<=i; j++) rumpname[j] = *(name+j);
   rumpname[j] = '\0';
}
   

/* allocate a type expression structure */ 
TEX *mktex(char *name)
{
   int     i;
   BOOL    found = FALSE;
   TEX *t = (TEX *) malloc(sizeof(TEX));
  
   /* initialize fields in order in which they are defined */

   stralloc(t->name, name);

   foreach_param(i) 
       t->kids[i] = NULL;

   foreach_attribute(i) {
       t->cpost[i].cpred    = DRC_TRUE;
       t->cpostres[i].cpred = DRC_TRUE;
   }

   t->parent     = NULL;
   t->childorder = 0;

   truncate(name);
   foreach_layer(i) {
      found = str_eq(layers[i].name,rumpname);
      if (found)
         break;
   }
   if ( !found ) {
      errorhead(ERR_MSG);
      error("Couldn't find component %s definition\n",name);
      exit(EXIT_FAILURE);
   }
   t->attrdata = &layers[i];
        
   t->next = NULL;

   return(t);
}

/* append child node to the end of the list lst */

TEX *append_sibling ( TEX *lst, TEX *node )
{
    TEX *p;
    int i;
    
    node->next = NULL;

    if ( lst == NULL ) { 
        node->childorder = 0;
        return(node);
    } 

    for(p=lst, i=1; p->next ; i++, p=p->next);
    p->next = node;
    node->childorder = i;
    return(lst);
}

/* test the number of parameters */

void test_num_params ( TEX *cdef, TEX *cinst )
{
    int  nt    = cdef->attrdata->num_param;
    TEX  *tmp;
    int  nkids;

    /* Step 1: count number of actual children (in nkids) */

    for (tmp = cinst, nkids = 0; tmp ; tmp =tmp->next, nkids++) ;

    /* Step 2: see if actual count equals defined count;
               if not, signal error */

    if (nt != nkids) { 
        errorhead(ERR_MSG);
        error("Error:   incorrect number of children associated with component %s\n",
              cdef->name);
        error("         %d expected but %d provided\n", nt, nkids );
    }
}

/* add all the kids to the parents & check their realms */
void add_kids ( TEX *parent, TEX *kidlist )
{
    TEX *p; 
    int i;

    for(p = kidlist, i=0; p ; p= p->next, i++) { 
        parent->kids[i] = p; 
        p->parent = parent;
        if (strcmp(p->attrdata->realmtype,
                   parent->attrdata->param_rtype[i])) { 
            errorhead(ERR_MSG);
            error("Realm type mismatch: ");
            error(" parameter #%d of %s should be from realm %s.\n", 
                  i+1, parent->name, parent->attrdata->param_rtype[i] );
            error("                      realm of %s is %s.\n",
                  p->name, p->attrdata->realmtype);
        }
    }
} 
   
/* return the union of the values in attributes cpost and post
   this is the "small-plus" operator of the DRC paper */

static void plus( COND *cpost, DRV post, TEX *t, COND *result )
{
   if ( post == DRC_INHERIT ) {
      result->cpred  = cpost->cpred;
      result->origin = cpost->origin;
   }
   else {
      result->cpred  = post;
      result->origin = t;
   }
}

void circle_plus( CONDVEC *cpost, DRV post[], TEX *t, CONDVEC *result )
{
   int i;

   foreach_attribute(i) {
      plus(&((*cpost)[i]), post[i], t,&((*result)[i]));
   }
}

/* returns TRUE if condition satisfies precondition;
   this is the small-sigma operation of the DRC paper  (formerly sat()) */

static BOOL sigma ( DRV pre, COND *cond )
{
   int condition = cond->cpred;

    switch ( pre ) { 
      case DRC_TRUE       : return( TRUE );  break;
      case DRC_ASSERT     : return ( condition == DRC_ASSERT ); break;
      case DRC_NEGATE     : return ( condition == DRC_NEGATE ); break;   
      case DRC_UNSET      : return ( condition == DRC_UNSET ); break;  
      case DRC_SET        : return ( condition == DRC_ASSERT || 
                                     condition == DRC_NEGATE ); break;  
      case DRC_NOT_NEGATE : return ( condition == DRC_ASSERT || 
                                     condition == DRC_UNSET ); break; 
      case DRC_NOT_ASSERT : return ( condition == DRC_NEGATE ||  
                                     condition == DRC_UNSET ); break;
      case DRC_FALSE      : return( FALSE ); break; 
    }
}

/* convert the attribute value into a string */
static char *value_word(DRV i)
{
    switch ( i ) { 
      case DRC_TRUE       : return("true"); break;
      case DRC_ASSERT     : return("assert"); break;
      case DRC_NEGATE     : return("negate"); break;
      case DRC_UNSET      : return("unset"); break;
      case DRC_SET        : return("assert or negate"); break;
      case DRC_NOT_NEGATE : return("assert or unset"); break;
      case DRC_NOT_ASSERT : return("negate or unset"); break;
      case DRC_FALSE      : return("false"); break;
      case DRC_INHERIT    : return("inherit"); break;
      default             : 
            printf("\n fatal error - can't recognize word\n");
            exit(EXIT_FAILURE);
    }
}



void satisfy( char *tname, DRV pre[], CONDVEC *cond, 
              int err_or_warn, int cond_or_res, int param )
{
   int  i;
   char *emsg    = (err_or_warn == ERR_MSG ? "error:  " : "warning:");
   char *msgtype = (cond_or_res == PRECOND_MSG ? "Precondition" : 
                     "Prerestriction");

   foreach_attribute(i) { 
      if (!sigma(pre[i],&(*cond)[i])) {
         errorhead(err_or_warn);

         /* print error or warning */
         error("%s %s\n   ", msgtype, emsg);

         if (cond_or_res == PRECOND_MSG) {
            if (pre[i] == DRC_ASSERT || 
                pre[i] == DRC_NOT_NEGATE ||
                pre[i] == DRC_SET ) {
               error("%s is expected ", attribute[i].desc);
               if ((*cond)[i].origin == NULL)
                  error("above %s\n", tname );
               else
                  error("between %s and %s\n", (*cond)[i].origin->name, tname);
            } 
            else {
               error("%s is not expected ", attribute[i].desc);
               if ((*cond)[i].origin == NULL)
                  error("above %s\n", tname );
               else
                  error("above %s\n", tname);
                  error("   upper layer %s caused inconsistency\n", 
                        (*cond)[i].origin->name);
            }
         }
         else {  /* prerestriction error */
            error("parameter %d of %s ", param+1, tname );
            if (pre[i] == DRC_ASSERT || 
                pre[i] == DRC_NOT_NEGATE ||
                pre[i] == DRC_SET )
               error("expects a subsystem with\n   %s ", attribute[i].desc);
            else
               error("does not expect a subsystem with\n   %s ", attribute[i].desc);

            if ((*cond)[i].origin == NULL)
               error("\n");
            else
               error("at or above %s\n", (*cond)[i].origin->name);
         }
      }
   }
}

/* performs the large SIGMA function of the DRC paper */
void Sigma( CONDVEC *cpost, DRV post[], TEX *t, CONDVEC *result )
{
   int i;
   foreach_attribute(i)
      plus( &((*cpost)[i]), post[i], t,&((*result)[i]) );
}

/* copy the attribute data from b to a */
void copyval(DRV a[], DRV b[])
{
    int i;
    for(i=0; i<MAXATTR; i++)
        a[i] = b[i];
}


void initialize_values( void )
{
   int i;
   foreach_attribute(i) {
      truevec[i].cpred  = DRC_TRUE;
      truevec[i].origin = NULL;
   }
}


/* determines if it is ok to merge the cumulative postrestrictions 
   of all the kids of component instance t. If it is ok, then the 
   value of the merge is returned in "oval" (output value). 
*/

void okmerge(TEX *t, CONDVEC *oval) 
{
    int i,j;
    int n0, n1, nA, nS;
    int nkids = t->attrdata->num_param;

    if ( nkids == 0 ) { 
        copycond( oval, &truevec ) ;
    } else if ( nkids == 1 ) { 
        copycond( oval, &t->kids[0]->cpostres ) ;
    } else { 
        for(j=0;j<num_of_attr; j++) { 
            n0 = 0; n1 = 0; nA = 0; nS = 0;
            for(i=0; i < nkids; i++ ) { 
                switch ( t->kids[i]->cpostres[j].cpred ) { 
                  case DRC_ASSERT : n1++; break;
                  case DRC_NEGATE : n0++; break;
                  case DRC_TRUE : nA++; break;
                  case DRC_UNSET : nS++; break;
                  default  : 
                      errorhead(ERR_MSG);
                      error("improper value %d for postrestriction for attribute %s",
                            t->kids[i]->cpostres[j], attribute[j].name);
                      num_te_errs++;
                      break;
                }
            }
            /* if all (nkids) have the same value, then 
               everything's ok and oval's attribute gets that value.
            */
            if ( n1 == nkids ) { 
                (*oval)[j].cpred  = DRC_ASSERT;
                (*oval)[j].origin = t->kids[0];
            } else if ( n0 == nkids ) { 
                (*oval)[j].cpred  = DRC_NEGATE;
                (*oval)[j].origin = t->kids[0];
            } else if ( nA == nkids ) {
                (*oval)[j].cpred  = DRC_TRUE;
                (*oval)[j].origin = t->kids[0];
            } else if ( nS == nkids ) {
                (*oval)[j].cpred  = DRC_UNSET;
                (*oval)[j].origin = t->kids[0];
            }
            /* if it's only DRC_TRUE and 1 other value type, then that's 
               okay too and the result is that other value. 
            */
              else if ( !n0 && !n1 ) { 
                (*oval)[j].cpred  = DRC_UNSET;
                (*oval)[j].origin = t->kids[0];
            } else if ( !n0 && !nS ) {
                (*oval)[j].cpred  = DRC_ASSERT;
                (*oval)[j].origin = t->kids[0];
            } else if ( !n1 && !nS ) { 
                (*oval)[j].cpred  = DRC_NEGATE;
                (*oval)[j].origin = t->kids[0];
            } else { 
                /* then it's not ok. print error message */
                errorhead(ERR_MSG);
                error("Incompatible merge of %s's children postrestrictions",t->name);
                error("For attribute |%s|, the results are:", attribute[j].name);
                for(i=0; i < nkids; i++ ) { 
                    int v = t->kids[i]->cpostres[j].cpred;
                    error("   kid[%d] (%s)'s value is %s\n",i, t->kids[i]->name, value_word(v));
                }
                (*oval)[j].cpred  = DRC_TRUE;
                (*oval)[j].origin = t->kids[0];
            }
        }
    }
}


/*********** stack routines ***************/

void push( CONDVEC *x )
{
   STACK *a = (STACK *) malloc( sizeof( STACK ) );
   copycond(&a->cpost,x);
   a->next = top;
   top = a;
}

void pop( void )
{
   STACK *a = top;
   top = top->next;
   free( (char*) a);
}

void get_top ( CONDVEC *x )
{  
   copycond(x,&top->cpost);
}

void pushcomponent( TEX *x )
{
   CSTACK *a = (CSTACK *) malloc( sizeof( CSTACK ) );
   a->te   = x;
   a->next = ctop;
   ctop    = a;
}

void popcomponent( void )
{
   CSTACK *a = ctop;
   ctop = ctop->next;
   free( (char*) a);
}

void get_ctop( TEX **x )
{ 
   *x = ctop->te;
}
  
void init_stacks()
{
   top  = NULL;
   init_current_conditions( NULL );
   push( &current_conditions );
   ctop = NULL;
}
 
/****************************************************

  MAIN PROGRAM 

*****************************************************/

static void error (char *format, ...)
{
  va_list       args;

  va_start(args, format);
  vfprintf(stdout, format, args);
  va_end(args);
}

static void pr_usage(char *fname)
{
      printf("Usage: %s [options] filename\n", fname );
      printf("       -a to display attribute values on standard output\n"); 
      printf("       -d for parser debugging\n"); 
      printf("       -[H|h] for this information\n");
      printf("       -f <drc_rule_files> to read rules from <drc_rule_files>\n");
      printf("          (filename default is \"%s\")\n",DEFAULT_ATTR_FILE);
      printf("       -w Write out attribute information as C code and then quit\n");
      printf("       -v Write out the version number of the program and then quit\n");
      printf("       -o Write out top-level value of type expressions\n");

}

main(argc, argv)
int  argc;
char **argv;
{
    
    int c;
    extern int optind, opterr;
    extern char *optarg;
    int attr_fname = 0;
    int wflag = 0;

    oflag = 0;
    aflag = 0;
    attdrcdebug = 0;
    tedrcdebug = 0;
    attdrc_flex_debug = 0;
    tedrc_flex_debug = 0;

    if (argc == 1) {
        pr_usage(argv[0]);
        exit(EXIT_SUCCESS);
    }
    
    while ((c = getopt (argc, argv, "adf:wvoHh")) != EOF) { 
        switch (c) { 
          case 'v' : 
              printf("Version #0.9, Monday 9/26/94\n");
              exit( EXIT_SUCCESS);
              break;
          case 'o' : oflag = 1; break;
          case 'w' : wflag = 1; break;
          case 'a' : aflag = 1; break;
          case 'd' : 
              attdrcdebug = 1; 
              tedrcdebug = 1; 
              attdrc_flex_debug = 1;
              tedrc_flex_debug = 1;
              break;
          case 'H' : 
          case 'h' : pr_usage(argv[0]); exit(EXIT_SUCCESS); break;
          case 'f' : 
              attr_fname = 1;
              attdrcin = fopen( optarg, "r");
              break;
        }
    }

    if ( !attr_fname ) 
        attdrcin = fopen( DEFAULT_ATTR_FILE , "r" );
    
    if (attdrcin == NULL) {
        printf( "Can't open attribute file\n" );
        exit(EXIT_FAILURE);
    }
    
    if (attdrcparse() != 0)
        printf( "attribute parsing failed\n");
    
    if ( wflag ) { 
        writeout();
        exit(EXIT_SUCCESS);
    }
    
    if (!( optind < argc) ) { 
        printf("missing file argument\n");
        exit(EXIT_FAILURE);
    }
    
    tedrcin = fopen ( argv[optind], "r");
    
    
    if (tedrcin == NULL) {
        printf( "Can't open type expression file\n" );
        exit(EXIT_FAILURE);
    }
    
    if ( tedrcparse() != 0)
        printf( "te parsing failed\n");

    exit(EXIT_SUCCESS);
}
