#define MAXATTR 15
#define MAXPARAM 3
#define MAXLAYER 60
#define MAXLENGTH 100           /* maximum length of a name */


typedef enum { 
               DRC_TRUE,        /*true                          */
               DRC_ASSERT,      /*assert                        */
               DRC_NEGATE,      /*negate                        */
               DRC_UNSET,       /*unset (not assert or negate)  */
               DRC_SET,         /*set (assert or negate)        */
               DRC_NOT_NEGATE,  /*not negate                    */
               DRC_NOT_ASSERT,  /*not assert                    */
               DRC_FALSE,       /*false                         */
               DRC_INHERIT      /*inherit                       */
} DRV;

#define DRC_INHERIT      8  /*inherit                       */

#define WARN_MSG 1
#define ERR_MSG 2

#define PRECOND_MSG  1
#define PRERES_MSG   2

#define EXIT_SUCCESS 0
#define EXIT_FAILURE 1



#define stralloc(X,Y)        { X = (char *) malloc(strlen(Y) + 1); strcpy(X,Y); }
#define str_eq(X,Y)          (strcmp(X,Y)==0)
#define foreach_attribute(i) for (i=0; i<num_of_attr; i++)
#define foreach_param(i)  for (i=0; i<MAXPARAM; i++)
#define foreach_layer(i)      for (i=0; i<=currlayer; i++)

extern int num_of_attr; /*actual number of attributes read in */

typedef enum { FALSE, TRUE } BOOL;

/* attribute definition structure */

typedef struct attr_def {
    char *name;                     /* name of attribute */
    char *desc;                     /* description of attribute */
    DRV  initvalue;                 /* initial value */
} ATTRDEF;
    
extern ATTRDEF attribute[MAXATTR]; 

/*component structure dealing with attributes */
typedef struct attr_str { 
    char *name;                     /*name of layer */
    DRV  pre[MAXATTR];              /*precondition */
    DRV  wpre[MAXATTR];             /*precondition warnings */
    DRV  post[MAXPARAM][MAXATTR];   /*postconditions */
    DRV  prer[MAXPARAM][MAXATTR];   /*prerestrictions */
    DRV  wprer[MAXPARAM][MAXATTR];  /*prerestrictions warnings */
    DRV  postr[MAXATTR];            /*postrestriction */
    BOOL copyval;                   /*Are postconditions copied 
                                       for all parameters? */
    BOOL mergeval;                  /*Are postrestrictions of children 
                                       merged? */
    char *realmtype;                /*realm of component */
    char *param_rtype[MAXPARAM];    /*realm of component's children */
    int  num_param;                 /*number of children for component */
} ATTRNODE;

struct tex;                         /*forward reference */

typedef struct {
    DRV         cpred;              /*cumulative predicate */
    struct tex  *origin;            /*component that originated pred */
} COND;

typedef COND  CONDVEC[MAXATTR];

/*type expression structure */
typedef struct tex { 
    char       *name;               /*name of component instance */
    struct tex *kids[MAXPARAM];     /*pointers to component's children */
    CONDVEC    cpost;               /*cumulative postcondition */
    CONDVEC    cpostres;            /*cumulative postrestriction */
    struct tex *parent;             /*pointer to component's parent */
    int        childorder;          /*this component is the (childorder)th
                                       child of its parent */
    ATTRNODE   *attrdata;           /*ptr to component attribute data */
    struct tex *next;               /*used to link kids together */
} TEX;

typedef struct stk {
   CONDVEC    cpost;
   struct stk *next;
} STACK;

typedef struct cstk {
   TEX         *te;
   struct cstk *next;
} CSTACK;

extern char *container_name;      /*the name of the type expression */
extern ATTRNODE layers[MAXLAYER]; /*the array of component attributes */

extern CONDVEC truevec;           /*attribute vector that's all "true" */

extern int num_te_errs;           /* # of errors for the type expression */
extern int num_te_msgs;           /* # of errors and warnings for the 
                                     type expression */
extern CONDVEC current_conditions;/* temporary used in validating
                                     preconditions */

/*attribute routines */
void addattr(char *name, char *desc, DRV value);
void add_to_lnames(char *s);
void setattrqual(BOOL mval, BOOL cpval);
void do_props(char *prop, int atype, int ctype, int param, int wtype);
void set_realm_type(char *realmname);
void addrparam(char *rp_name);

/*type expression routines */

void copycond( CONDVEC *x, CONDVEC *y );
void setcpostres( CONDVEC *x, TEX *t );
void initialize_values(void);
void set_typex_name(char *name);
TEX *mktex(char *name);
TEX *append_sibling(TEX *a, TEX *b);
void add_kids(TEX *parent, TEX *kidlist);
void satisfy( char *tname, DRV pre[], CONDVEC *cond,
              int err_or_warn, int cond_or_res, int param );
void circle_plus( CONDVEC *cpost, DRV post[], TEX *t, CONDVEC *result );
void okmerge(TEX *t, CONDVEC *oval);
void copyval(DRV a[], DRV b[]);
void test_num_params(TEX *t, TEX *kids);
void print_top_te_values(TEX *t);
void init_stacks( void );
