/* type equation grammar (bison source file). */

%{

#include "d.h"

%}

/************************************************************************/
/*              Tokens.                                                 */
/************************************************************************/
%token NAME

%union {
  char   str[100];
  TEX    *nptr;
}
%type <str>  NAME
%type <nptr> te_list te_expr;

%start te_gram

%%

te_gram: {
          initialize_values();
         }
         teexprs;

teexprs: testmt | teexprs testmt ;


testmt : NAME '=' 
         { 
           set_typex_name($1);
           init_stacks();
           num_te_errs = 0;
           num_te_msgs = 0;
         }
         te_expr ';'
         {  
           if ( num_te_errs == 0 )
               printf(" >>>> %s has 0 errors, %d warnings <<<<\n",
                  $1, num_te_msgs-num_te_errs);
           else
               printf("=============\n");
           print_top_te_values($4); 
         }
       ;

te_expr:  NAME 
         {
           TEX *t = mktex($1);

           /* test preconditions */
           get_top(&current_conditions);
           satisfy( t->name, t->attrdata->pre, &current_conditions, 
                    ERR_MSG, PRECOND_MSG, 0);
           satisfy( t->name, t->attrdata->wpre, &current_conditions, 
                    WARN_MSG, PRECOND_MSG, 0);

           /* no need to test prerestrictions as there
              are none */

           /* set up for postrestrictions */
           circle_plus( &truevec, t->attrdata->postr, t, &t->cpostres );

           /* before we leave, test if we got the right number of 
              parameters */
           test_num_params(t, NULL);

           pop();
           $$ = t;
         }

       |  NAME 
         {
           int i;
           TEX *t;
           COND tmp[MAXATTR];

           t = mktex($1);

           /* test preconditions */
           get_top(&current_conditions);
           satisfy( t->name, t->attrdata->pre, &current_conditions, 
                    ERR_MSG, PRECOND_MSG, 0);
           satisfy( t->name, t->attrdata->wpre, &current_conditions, 
                    WARN_MSG, PRECOND_MSG, 0);

           /* we need to push conditions onto the stack backwards,
              so the first child is on top */

           for (i=t->attrdata->num_param-1; i>=0; i--) { 

              /* propagate current conditions to first parameter
                 and also to other parameters if merging occurs */

              if (i == 0 || t->attrdata->mergeval)
                 circle_plus(&current_conditions, t->attrdata->post[i], 
                             t, &tmp);
              else {
                 init_current_conditions( t );
                 circle_plus(&current_conditions, t->attrdata->post[i], 
                             t, &tmp);
              }
              push(tmp);
           }

           /* add this component to the top of the stack to pop off
              after parsing te_list */

           pushcomponent( t );
         }
          '[' te_list  ']'
         {
           TEX *t;
           int  i;
           CONDVEC tmp;

           get_ctop(&t);

           /* see if the right number of parameters are present
              and add the "kids" to the node */

           test_num_params(t, $4);
           add_kids(t, $4);

           /* test prerestrictions */

           for (i=0; i<t->attrdata->num_param; i++) {
              satisfy( t->name, t->attrdata->prer[i], &t->kids[i]->cpostres,
                       ERR_MSG, PRERES_MSG, i);
              satisfy( t->name, t->attrdata->wprer[i], &t->kids[i]->cpostres, 
                       WARN_MSG, PRERES_MSG, i);
           }

           /* now compute cumulative postrestrictions for component */

           if ( t->attrdata->mergeval ) {
              okmerge(t,&tmp);
              circle_plus(&tmp, t->attrdata->postr, t, &t->cpostres);
           }
           else {
              circle_plus(&t->kids[0]->cpostres, t->attrdata->postr, t,
                          &t->cpostres);
           }
              
           /* remove top and ctop */
           pop();
           popcomponent();

           $$ = t;
          }
       ;

te_list: te_expr              { $$ = append_sibling(NULL,$1); }
       | te_list ',' te_expr  { $$ = append_sibling($1, $3);  }
       ;

%%

int tedrcwrap (void)
{
  return(1);
}

int tedrcerror (char *s)
{
  return 0;
}

